import * as mc from "@minecraft/server"
import { world, system } from "@minecraft/server"
import CommandBuilder from "./Modules/CommandBuilder"
import Restful from "./Modules/Restful"

const WandID = "minecraft:wooden_axe"
const UndoStack = {}
const RedoStack = {}
const Clipboard = {}

/**
 * @param {mc.Vector3} location1
 * @param {mc.Vector3} location2
 */
const CalculateSize = (location1, location2) => {
    const volume = new mc.BlockVolume(location1, location2)
    return volume.getCapacity()
}
const Pos1 = {}
/**
 * @param {mc.Player} player
 */
const SetPos1 = (player, location) => {
    if (Date.now() < (player.getDynamicProperty("Pos1_Cooldown") ?? Date.now())) return
    player.setDynamicProperty("Pos1_Cooldown", Date.now() + 500)
    Pos1[player.name] = location
    let size = undefined
    if (Pos2[player.name] != undefined) size = CalculateSize(Pos1[player.name], Pos2[player.name])
    player.sendMessage(`§dFirst position set to (${Object.values(location).map(l => Math.floor(l).toFixed(1)).join(", ")}) ${size ? `(${size})` : ""}`)
}
const Pos2 = {}
/**
 * @param {mc.Player} player
 */
const SetPos2 = (player, location) => {
    if (Date.now() < (player.getDynamicProperty("Pos2_Cooldown") ?? Date.now())) return
    player.setDynamicProperty("Pos2_Cooldown", Date.now() + 500)
    Pos2[player.name] = location
    let size = undefined
    if (Pos1[player.name] != undefined) size = CalculateSize(Pos1[player.name], Pos2[player.name])
    player.sendMessage(`§dSecond position set to (${Object.values(location).map(l => Math.floor(l).toFixed(1)).join(", ")}) ${size ? `(${size})` : ""}`)
}

const PositionError = (player) => {
    if (Pos1[player.name] == undefined || Pos2[player.name] == undefined) return true
    return false
}

/**
 * @param {string} blockId
 */
const GetBlock = (blockId) => {
    return mc.BlockTypes.get(blockId)
}

/**
 * @param {mc.Player} player
 */
const IsAdmin = (player) => {
    return player.getTags().find(t => t.toLowerCase() == "admin" || t.toLowerCase() == "admin:worldedit") != undefined
}

const warn = console.warn
console.warn = (...args) => {
    warn("[Simple World Edit]", ...args)
}

/**
 * @param {mc.Entity} entity
 * @returns {mc.ItemStack}
 */
const getMainhand = (entity) => {
    return entity.getComponent("equippable")?.getEquipment(mc.EquipmentSlot.Mainhand)
}

world.afterEvents.worldInitialize.subscribe(() => {
    console.warn("Loaded.")
})

world.beforeEvents.playerBreakBlock.subscribe((data) => {
    const { player, itemStack, block } = data
    if (!IsAdmin(player)) return
    if (itemStack?.typeId === WandID) {
        data.cancel = true
        system.run(() => SetPos1(player, block.location))
    }
})

world.afterEvents.entityHitBlock.subscribe(({ damagingEntity: player, hitBlock: block }) => {
    if (!IsAdmin(player)) return
    if (getMainhand(player)?.typeId === WandID) {
        SetPos1(player, block.location)
    }
}, { entityTypes: ["minecraft:player"] })

world.beforeEvents.playerInteractWithBlock.subscribe((data) => {
    const { player, itemStack, block } = data
    if (!IsAdmin(player)) return
    if (itemStack?.typeId === WandID) {
        data.cancel = true
        system.run(() => SetPos2(player, block.location))
    }
})

world.afterEvents.playerDimensionChange.subscribe(({ player }) => {
    Pos1[player.name] = undefined
    Pos2[player.name] = undefined
})

/**
 * @param {mc.Player} player
 * @param {mc.ListBlockVolume} volume
 */
const AddUndo = (player, volume) => {
    if (UndoStack[player.name] == undefined) UndoStack[player.name] = []
    const blocks = []
    for (const location of volume.getBlockLocationIterator()) {
        const block = player.dimension.getBlock(location)
        blocks.push({ location: location, dimension: player.dimension, block: block.permutation })
    }
    UndoStack[player.name].push(blocks)
}

/**
 * @param {mc.Player} player
 * @returns {{location: mc.Vector3, dimension: mc.Dimension, block: mc.BlockPermutation}[]}
 */
const GetUndo = (player) => {
    if (UndoStack[player.name] == undefined) UndoStack[player.name] = []
    return UndoStack[player.name].pop()
}

/**
 * @param {mc.Player} player
 * @param {mc.ListBlockVolume} volume
 */
const AddRedo = (player, volume) => {
    if (RedoStack[player.name] == undefined) RedoStack[player.name] = []
    const blocks = []
    for (const location of volume.getBlockLocationIterator()) {
        const block = player.dimension.getBlock(location)
        blocks.push({ location: location, dimension: player.dimension, block: block.permutation })
    }
    RedoStack[player.name].push(blocks)
}

/**
 * @param {mc.Player} player
 * @returns {{location: mc.Vector3, dimension: mc.Dimension, block: mc.BlockPermutation}[]}
 */
const GetRedo = (player) => {
    if (RedoStack[player.name] == undefined) RedoStack[player.name] = []
    return RedoStack[player.name].pop()
}

/**
 * @param {mc.Player} player
 * @returns {{location: mc.Vector3, dimension: mc.Dimension, block: mc.BlockPermutation}[]}
 */
const ClearRedo = (player) => {
    if (RedoStack[player.name] != undefined) RedoStack[player.name] = []
}

/**
 * @param {mc.Vector3} location 
 * @param {number} angle 
 */
const RotateLocation = (location, angle) => {
    const { x, y, z } = location
    switch (angle) {
        case 90:
            return { x: -z, y, z: x }
        case 180:
            return { x: -x, y, z: -z }
        case 270:
            return { x: z, y, z: -x }
        default:
            return { x, y, z }
    }
}

const Command = new CommandBuilder("simpleworldedit")
world.beforeEvents.chatSend.subscribe(async (data) => {
    let player = data.sender
    if (!IsAdmin(player)) return
    let message = data.message
    let prefix = "!"
    if (message.startsWith(prefix)) {
        let args = data.message.slice(prefix.length).trim().split(' ')
        let command = args.shift().toLowerCase()
        let cmd = Command.getAllRegistation().find(c => c.name == command || (c.aliases && c.aliases.includes(command)))
        data.cancel = true
        if (!cmd)
            return player.sendMessage(`§cUnknown command: ${command}. Please check that the command exists and that you have permission to use it.`)
        system.run(() => {
            player.runCommandAsync(`scriptevent ${cmd.id}:runCommand ${command} ${args.join(" ")}`)
            if (command != "undo") ClearRedo(player)
        })
    }
})

Command.afterCommandUse.subscribe(({ error, player, command }) => {
    if (command.id != "simpleworldedit") return
    if (command.name != "undo") ClearRedo(player)
    if (error) {
        player.sendMessage(`§c${error}${error?.stack ?? ""}`)
    }
})

Command.register({
    name: "pos1",
    description: "Select the player's bottom half as position 1",
    category: "World Edit",
    permission: "worldedit",
    usage: "pos1"
},
    (data, player, args) => {
        SetPos1(player, player.location)
    })

Command.register({
    name: "pos2",
    description: "Select the player's bottom half as position 2",
    category: "World Edit",
    permission: "worldedit",
    usage: "pos2"
},
    (data, player, args) => {
        SetPos2(player, player.location)
    })

Command.register({
    name: "hpos1",
    description: "Select the block you are looking at as position 2",
    category: "World Edit",
    permission: "worldedit",
    usage: "hpos1"
},
    (data, player, args) => {
        const block = player.getBlockFromViewDirection()?.block
        if (!block) return player.sendMessage("§cYou have to look at the block.")
        SetPos1(player, block.location)
    })

Command.register({
    name: "hpos2",
    description: "Select the block you are looking at as position 2",
    category: "World Edit",
    permission: "worldedit",
    usage: "hpos2"
},
    (data, player, args) => {
        const block = player.getBlockFromViewDirection()?.block
        if (!block) return player.sendMessage("§cYou have to look at the block.")
        SetPos2(player, block.location)
    })

Command.register({
    name: "set",
    description: "Set all blocks from the region to selected block",
    category: "World Edit",
    permission: "worldedit",
    usage: "set <to-block>"
},
    (data, player, args) => {
        if (PositionError(player)) return player.sendMessage("§cYou have to select the area.")
        const id = args[0]
        if (!id) return player.sendMessage("§cInput the block id.")
        const targetBlock = GetBlock(id)

        const volume = new mc.BlockVolume(Pos1[player.name], Pos2[player.name])
        AddUndo(player, player.dimension.getBlocks(
            volume,
            {},
            true
        ))

        let blocksChanged = 0
        for (const location of volume.getBlockLocationIterator()) {
            if (player.dimension.getBlock(location) == undefined) continue
            player.dimension.runCommand(`setblock ${Object.values(location).join(" ")} ${targetBlock.id}`)
            blocksChanged++
        }

        player.sendMessage(`§d${blocksChanged} blocks(s) have been changed.`)
    })

Command.register({
    name: "replace",
    description: "Replace all block type in region to selected block",
    category: "World Edit",
    permission: "worldedit",
    usage: "replace <from-block> <to-block>"
},
    (data, player, args) => {
        if (PositionError(player)) return player.sendMessage("§cYou have to select the area.")
        const replace = args[0]
        if (!replace) return player.sendMessage("§cInput the block id.")
        const id = args[1]

        let targetId = replace
        let replaceId = undefined
        if (id) {
            targetId = id
            replaceId = replace
        }

        const targetBlock = GetBlock(targetId)
        const replaceBlock = GetBlock(replaceId)

        const volume = new mc.BlockVolume(Pos1[player.name], Pos2[player.name])
        AddUndo(player, player.dimension.getBlocks(
            volume,
            {},
            true
        ))

        let blocksChanged = 0
        for (const location of volume.getBlockLocationIterator()) {
            if (player.dimension.getBlock(location) == undefined) continue
            if (replaceBlock && (player.dimension.getBlock(location).typeId != replaceBlock.id)) continue
            player.dimension.runCommand(`setblock ${Object.values(location).join(" ")} ${targetBlock.id}`)
            blocksChanged++
        }

        player.sendMessage(`§d${blocksChanged} blocks(s) have been changed.`)
    })

Command.register({
    name: "copy",
    description: "Copy the selected region. And your position as reference",
    category: "World Edit",
    permission: "worldedit",
    usage: "copy"
},
    (data, player, args) => {
        if (PositionError(player)) return player.sendMessage("§cYou have to select the area.")
        const clipData = []

        let blocksCopied = 0
        const volume = new mc.BlockVolume(Pos1[player.name], Pos2[player.name])
        for (const location of volume.getBlockLocationIterator()) {
            const block = player.dimension.getBlock(location)
            if (!block) continue
            const newLocation = {
                x: block.location.x - Math.floor(player.location.x),
                y: block.location.y - Math.floor(player.location.y),
                z: block.location.z - Math.floor(player.location.z)
            }
            clipData.push({ location: newLocation, block: block.permutation })
            blocksCopied++
        }
        Clipboard[player.name] = {
            blocks: clipData,
            dimension: player.dimension,
            angle: 0
        }

        player.sendMessage(`§d${blocksCopied} block(s) were copied.`)
    })

Command.register({
    name: "paste",
    description: "Paste the contents of the clipboard",
    category: "World Edit",
    permission: "worldedit",
    usage: "paste"
},
    (data, player, args) => {

        const clipData = Clipboard[player.name]
        if (!clipData) return player.sendMessage("§cYou have no block copied.")
        const undoData = []

        for (const data of clipData.blocks) {
            const rotatedLocation = RotateLocation(data.location, clipData.angle)
            const location = {
                x: player.location.x + rotatedLocation.x,
                y: player.location.y + rotatedLocation.y,
                z: player.location.z + rotatedLocation.z
            }
            const block = clipData.dimension.getBlock(location)
            if (!block) continue
            undoData.push({
                location: block.location,
                dimension: block.dimension,
                block: block.permutation
            })

            clipData.dimension.setBlockPermutation(location, data.block)
        }

        if (UndoStack[player.name] == undefined) UndoStack[player.name] = []
        UndoStack[player.name].push(undoData)

        player.sendMessage(`§dThe clipboard has been pasted at (${Object.values(player.location).map(l => Math.floor(l).toFixed(1)).join(", ")})`)
    })

Command.register({
    name: "rotate",
    description: "Rotate the contents of the clipboard",
    category: "World Edit",
    permission: "worldedit",
    usage: "rotate <angle: 0/90/180/270>"
},
    (data, player, args) => {

        const clipData = Clipboard[player.name]
        if (!clipData) return player.sendMessage("§cYou have no block copied.")
        if (!args[0]) return player.sendMessage("§cInput the angle.")
        const angle = Number(args[0])
        const allowedAngle = [0, 90, 180, 270]
        if (!allowedAngle.includes(angle)) return player.sendMessage("§cInput the angle.")

        clipData.angle = angle

        player.sendMessage(`§dThe clipboard has been rotated to ${angle} degree`)
    })

Command.register({
    name: "count",
    description: "Count the number of blocks in the region",
    category: "World Edit",
    permission: "worldedit",
    usage: "count"
},
    (data, player, args) => {
        if (PositionError(player)) return player.sendMessage("§cYou have to select the area.")

        const blockId = args[0]
        let blocks = 0
        const volume = new mc.BlockVolume(Pos1[player.name], Pos2[player.name])
        for (const location of volume.getBlockLocationIterator()) {
            const block = player.dimension.getBlock(location)
            if (block == undefined) continue
            if (blockId) {
                if (block.typeId == blockId) blocks++
            } else {
                if (block.isSolid) blocks++
            }
        }

        player.sendMessage(`§d${blocks} block(s) in the selected area.`)
    })

Command.register({
    name: "undo",
    description: "Removes your last action",
    category: "World Edit",
    permission: "worldedit",
    usage: "undo"
},
    (data, player, args) => {
        const redoData = []
        const undo = GetUndo(player)
        if (!undo) return
        for (const data of undo) {
            if (data.dimension.getBlock(data.location) == undefined) continue
            redoData.push({
                location: data.location,
                dimension: data.dimension,
                block: data.dimension.getBlock(data.location).permutation
            })
            data.dimension.setBlockPermutation(data.location, data.block)
        }
        if (RedoStack[player.name] == undefined) RedoStack[player.name] = []
        RedoStack[player.name].push(redoData)

        player.sendMessage("§dUndo successful.")
    })

Command.register({
    name: "redo",
    description: "Resets the last removed action",
    category: "World Edit",
    permission: "worldedit",
    usage: "redo"
},
    (data, player, args) => {
        const undoData = []
        const redo = GetRedo(player)
        if (!redo) return
        for (const data of redo) {
            if (data.dimension.getBlock(data.location) == undefined) continue
            undoData.push({
                location: data.location,
                dimension: data.dimension,
                block: data.dimension.getBlock(data.location).permutation
            })
            data.dimension.setBlockPermutation(data.location, data.block)
        }
        if (UndoStack[player.name] == undefined) UndoStack[player.name] = []
        UndoStack[player.name].push(undoData)

        player.sendMessage("§dRedo successful.")
    })

world.afterEvents.worldInitialize.subscribe((data) => {
    Restful.request("mce-installed").then((data) => {
        if (!data?.installed) {
            console.warn("Minecraft Essentials not detected.")
            Command.register({
                name: "help",
                description: "Provides help/list of world edit commands",
                category: "World Edit",
                permission: "worldedit",
                usage: "help"
            },
                /**
                 * @param {mc.Player} player
                 * @param {string[]} args
                 */
                (data, player, args) => {
                    const commandList = Command.getAllRegistation().filter(c => c.id == "simpleworldedit")
                    let helpMessage = `§eHelp Message ("?" Its means optional or not must be filled)`
                    helpMessage += `\n§aWorld Edit Command :`
                    for (const command of commandList) {
                        const commandName = "!" + command.name
                        const commandUsage = "!" + command.usage
                        helpMessage += `\n§e - §a${commandName} §e| ${commandUsage} (${command.description})`
                    }
                    player.sendMessage(helpMessage)
                })
        }
    })
})